@@ -19,7 +19,7 @@ use warnings;
 
 package MongoDB;
 {
-  $MongoDB::VERSION = '0.46.2';
+  $MongoDB::VERSION = '0.46_3';
 }
 # ABSTRACT: A Mongo Driver for Perl
 
@@ -64,7 +64,6 @@ through the MongoDB IRC channel, I<irc.freenode.net#mongodb>.
 
   Florian Ragwitz <rafl@debian.org>
   Kristina Chodorow <kristina@mongodb.org>
-  Mike Friedman <mike.friedman@10gen.com>
 
 =head1 COPYRIGHT AND LICENSE
 
