@@ -18,7 +18,7 @@
 #include "perl_mongo.h"
 
 static int mongo_link_sockaddr(struct sockaddr_in *addr, char *host, int port);
-static int mongo_link_reader(int socket, void *dest, int len);
+static int mongo_link_reader(mongo_link* link, void *dest, int len);
 
 /**
  * Waits "timeout" ms for the socket to be ready.  Returns 1 on success, 0 on
@@ -26,13 +26,43 @@ static int mongo_link_reader(int socket, void *dest, int len);
  */
 static int mongo_link_timeout(int socket, time_t timeout);
 
+static void set_timeout(int socket, time_t timeout) {
+#ifdef WIN32
+  DWORD tv = (DWORD)timeout * 1000;
+  const char *tv_ptr = (const char*)&tv;
+#else
+  struct timeval tv;
+  tv.tv_sec = 1;
+  tv.tv_usec = 0;
+  const void *tv_ptr = (void*)&tv;
+#endif
+  setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, tv_ptr, sizeof(tv));
+  setsockopt(socket, SOL_SOCKET, SO_SNDTIMEO, tv_ptr, sizeof(tv));
+}
+
+
+void perl_mongo_connect(mongo_link* link) {
+#ifdef MONGO_SSL
+  if(link->ssl){
+    ssl_connect(link);
+    link->sender = ssl_send;
+    link->receiver = ssl_recv;
+    return;
+  }
+#endif
+
+  non_ssl_connect(link);
+  link->sender = non_ssl_send;
+  link->receiver = non_ssl_recv;
+}
+
 /*
  * Returns -1 on failure, the socket fh on success.
  *
  * Note: this cannot return 0 on failure, because reconnecting sometimes makes
  * the fh 0 (briefly).
  */
-int perl_mongo_connect(char *host, int port, int timeout) {
+void non_ssl_connect(mongo_link* link) {
   int sock, status, connected = 0;
   struct sockaddr_in addr;
 
@@ -47,13 +77,13 @@ int perl_mongo_connect(char *host, int port, int timeout) {
   error = WSAStartup(version, &wsaData);
 
   if (error != 0) {
-    return -1;
+    return;
   }
 
   // create socket
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (sock == INVALID_SOCKET) {
-    return -1;
+    return;
   }
 
 #else
@@ -62,17 +92,18 @@ int perl_mongo_connect(char *host, int port, int timeout) {
   // create socket
   if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
     croak("couldn't create socket: %s\n", strerror(errno));
-    return -1;
+    return;
   }
 #endif
 
   // get addresses
-  if (!mongo_link_sockaddr(&addr, host, port)) {
-    return -1;
+  if (!mongo_link_sockaddr(&addr, link->master->host, link->master->port)) {
+    return;
   }
 
   setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &yes, INT_32);
   setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &yes, INT_32);
+  set_timeout(sock, link->timeout);
 
 #ifdef WIN32
   ioctlsocket(sock, FIONBIO, (u_long*)&yes);
@@ -94,18 +125,18 @@ int perl_mongo_connect(char *host, int port, int timeout) {
     if (errno != EINPROGRESS)
 #endif
     {
-      return -1;
+      return;
     }
 
-    if (!mongo_link_timeout(sock, timeout)) {
-      return -1;
+    if (!mongo_link_timeout(sock, link->timeout)) {
+      return;
     }
 
     size = sizeof(addr);
 
     connected = getpeername(sock, (struct sockaddr*)&addr, &size);
-    if (connected == -1) {
-      return -1;
+    if (connected == -1){
+      return;
     }
   }
   else if (status == 0) {
@@ -118,7 +149,67 @@ int perl_mongo_connect(char *host, int port, int timeout) {
 #else
   fcntl(sock, F_SETFL, 0);
 #endif
-  return sock;
+  link->master->socket = sock;
+  link->master->connected = 1;
+
+  return;
+}
+
+#ifdef MONGO_SSL
+// Establish a connection using an SSL layer
+void ssl_connect(mongo_link* link) {
+  tcp_setup(link);
+
+  if (link->master->socket){
+    // Register the error strings for libcrypto & libssl
+    SSL_load_error_strings();
+
+    // Register the available ciphers and digests
+    SSL_library_init();
+
+    // New context saying we are a client, and using SSL 2 or 3
+    link->ssl_context = SSL_CTX_new(SSLv23_client_method());
+    if(link->ssl_context == NULL){
+      ERR_print_errors_fp(stderr);
+    }
+
+    // Create an SSL struct for the connection
+    link->ssl_handle = SSL_new(link->ssl_context);
+    if(link->ssl_handle == NULL){
+      ERR_print_errors_fp(stderr);
+    }
+
+    // Connect the SSL struct to our connection
+    if(!SSL_set_fd(link->ssl_handle, link->master->socket)){
+      ERR_print_errors_fp(stderr);
+    }
+
+    // Initiate SSL handshake
+    if(SSL_connect (link->ssl_handle) != 1){
+      ERR_print_errors_fp(stderr);
+    }
+
+    SSL_CTX_set_timeout(link->ssl_context, (long)link->timeout);
+
+    link->master->connected = 1;
+  }
+}
+
+int ssl_send(void* link, const char* buffer, size_t len){
+  return SSL_write(((mongo_link*)link)->ssl_handle, buffer, len);
+}
+
+int ssl_recv(void* link, const char* buffer, size_t len){
+  return SSL_read(((mongo_link*)link)->ssl_handle, (void*)buffer, len);
+}
+#endif
+
+int non_ssl_send(void* link, const char* buffer, size_t len){
+  return send(((mongo_link*)link)->master->socket, buffer, len, 0);
+}
+
+int non_ssl_recv(void* link, const char* buffer, size_t len){
+  return recv(((mongo_link*)link)->master->socket, (void*)buffer, len, 0);
 }
 
 static int mongo_link_timeout(int sock, time_t to) {
@@ -128,7 +219,7 @@ static int mongo_link_timeout(int sock, time_t to) {
     return 1;
   }
 
-  timeout.tv_sec = to > 0 ? (to / 1000) : 20;
+  timeout.tv_sec = to > 0 ? ((long)to / 1000) : 20;
   timeout.tv_usec = to > 0 ? ((to % 1000) * 1000) : 0;
 
   // initialize prev, in case we get interrupted
@@ -222,27 +313,34 @@ static int mongo_link_sockaddr(struct sockaddr_in *addr, char *host, int port) {
  */
 int mongo_link_say(SV *link_sv, buffer *buf) {
   int sock, sent;
+  mongo_link *link;
+  link = (mongo_link*)perl_mongo_get_ptr_from_instance(link_sv, &connection_vtbl);
 
   if ((sock = perl_mongo_master(link_sv, 1)) == -1) {
     return -1;
   }
-  
-  sent = send(sock, (const char*)buf->start, buf->pos-buf->start, 0);
-  
+
+  sent = link->sender(link, (const char*)buf->start, buf->pos-buf->start);
+
   if (sent == -1) {
     set_disconnected(link_sv);
   }
-  
+
   return sent;
 }
 
 
 static int get_header(int sock, SV *cursor_sv, SV *link_sv) {
   mongo_cursor *cursor;
+  mongo_link *link;
+  int size;
 
   cursor = (mongo_cursor*)perl_mongo_get_ptr_from_instance(cursor_sv, &cursor_vtbl);
+  link = (mongo_link*)perl_mongo_get_ptr_from_instance(link_sv, &connection_vtbl);
+  size = 0;
 
-  if (recv(sock, (char*)&cursor->header.length, INT_32, 0) != INT_32) {
+  size = link->receiver(link, (char*)&cursor->header.length, INT_32);
+  if(size != INT_32){
     set_disconnected(link_sv);
     return 0;
   }
@@ -250,16 +348,14 @@ static int get_header(int sock, SV *cursor_sv, SV *link_sv) {
   cursor->header.length = MONGO_32(cursor->header.length);
 
   // make sure we're not getting crazy data
-  if (cursor->header.length > MAX_RESPONSE_LEN ||
-      cursor->header.length < REPLY_HEADER_SIZE) {
-
+  if (cursor->header.length > MAX_RESPONSE_LEN || cursor->header.length < REPLY_HEADER_SIZE) {
     set_disconnected(link_sv);
     return 0;
   }
 
-  if (recv(sock, (char*)&cursor->header.request_id, INT_32, 0) != INT_32 ||
-      recv(sock, (char*)&cursor->header.response_to, INT_32, 0) != INT_32 ||
-      recv(sock, (char*)&cursor->header.op, INT_32, 0) != INT_32) {
+  if (link->receiver(link, (char*)&cursor->header.request_id, INT_32)  != INT_32  ||
+      link->receiver(link, (char*)&cursor->header.response_to, INT_32) != INT_32  ||
+      link->receiver(link, (char*)&cursor->header.op, INT_32)          != INT_32) {
     return 0;
   }
 
@@ -333,7 +429,7 @@ int mongo_link_hear(SV *cursor_sv) {
       return 0;
     }
 
-    if (recv(sock, (char*)temp, 20, 0) == -1) {
+    if (link->receiver(link, (char*)temp, 20) == -1) {
       SvREFCNT_dec(link_sv);
       SvREFCNT_dec(request_id_sv);
       croak("couldn't get header response to throw out");
@@ -344,7 +440,7 @@ int mongo_link_hear(SV *cursor_sv) {
       int temp_len = len > 4096 ? 4096 : len;
       len -= temp_len;
 
-      if (mongo_link_reader(sock, (void*)temp, temp_len) == -1) {
+      if (mongo_link_reader(link, (void*)temp, temp_len) == -1) {
         SvREFCNT_dec(link_sv);
         SvREFCNT_dec(request_id_sv);
         croak("couldn't get response to throw out");
@@ -355,15 +451,16 @@ int mongo_link_hear(SV *cursor_sv) {
     if (get_header(sock, cursor_sv, link_sv) == 0) {
       SvREFCNT_dec(link_sv);
       SvREFCNT_dec(request_id_sv);
+      croak("invalid header received");
       return 0;
     }
   }
   SvREFCNT_dec(request_id_sv);
-  
-  if (recv(sock, (char*)&cursor->flag, INT_32, 0) == -1 ||
-      recv(sock, (char*)&cursor->cursor_id, INT_64, 0) == -1 ||
-      recv(sock, (char*)&cursor->start, INT_32, 0) == -1 ||
-      recv(sock, (char*)&num_returned, INT_32, 0) == -1) {
+
+  if (link->receiver(link, (char*)&cursor->flag, INT_32)      == -1 ||
+      link->receiver(link, (char*)&cursor->cursor_id, INT_64) == -1 ||
+      link->receiver(link, (char*)&cursor->start, INT_32)     == -1 ||
+      link->receiver(link, (char*)&num_returned, INT_32)      == -1) {
     SvREFCNT_dec(link_sv);
     croak("%s", strerror(errno));
     return 0;
@@ -389,38 +486,38 @@ int mongo_link_hear(SV *cursor_sv) {
     Newx(cursor->buf.start, cursor->header.length, char);
     cursor->buf.end = cursor->buf.start + cursor->header.length;
   }
-  else if (cursor->buf.end - cursor->buf.start < cursor->header.length) { 
+  else if (cursor->buf.end - cursor->buf.start < cursor->header.length) {
     Renew(cursor->buf.start, cursor->header.length, char);
     cursor->buf.end = cursor->buf.start + cursor->header.length;
   }
   cursor->buf.pos = cursor->buf.start;
 
-  if (mongo_link_reader(sock, cursor->buf.pos, cursor->header.length) == -1) {
+  if (mongo_link_reader(link, cursor->buf.pos, cursor->header.length) == -1) {
 #ifdef WIN32
     croak("WSA error getting database response: %d\n", WSAGetLastError());
 #else
     croak("error getting database response: %s\n", strerror(errno));
-#endif 
+#endif
     return 0;
   }
-  
+
   cursor->num += num_returned;
   return num_returned > 0;
 }
- 
+
 
 /*
  * Low-level func to get a response from the MongoDB server
  */
-static int mongo_link_reader(int socket, void *dest, int len) {
+static int mongo_link_reader(mongo_link* link, void *dest, int len) {
   int num = 1, read = 0;
 
   // this can return FAILED if there is just no more data from db
-  while(read < len && num > 0) {
+  while (read < len && num > 0) {
     int temp_len = (len - read) > 4096 ? 4096 : (len - read);
 
     // windows gives a WSAEFAULT if you try to get more bytes
-    num = recv(socket, (char*)dest, temp_len, 0);
+    num = link->receiver(link, (char*)dest, temp_len);
 
     if (num < 0) {
       return -1;
@@ -454,6 +551,12 @@ void set_disconnected(SV *link_sv) {
   close(link->master->socket);
 #endif
 
+#ifdef MONGO_SSL
+  if(link->ssl){
+    ssl_disconnect(link);
+  }
+#endif
+
   link->master->connected = 0;
 
   // TODO: set $self->_master to 0?
@@ -492,6 +595,13 @@ int perl_mongo_master(SV *link_sv, int auto_reconnect) {
     m_link = (mongo_link*)perl_mongo_get_ptr_from_instance(master, &connection_vtbl);
     link->copy = 1;
     link->master = m_link->master;
+    link->ssl = m_link->ssl;
+#ifdef MONGO_SSL
+    link->ssl_handle = m_link->ssl_handle;
+    link->ssl_context = m_link->ssl_context;
+#endif
+    link->sender = m_link->sender;
+    link->receiver = m_link->receiver;
 
     return link->master->socket;
   }
@@ -499,3 +609,42 @@ int perl_mongo_master(SV *link_sv, int auto_reconnect) {
   link->master = 0;
   return -1;
 }
+
+#ifdef MONGO_SSL
+// Establish a regular tcp connection
+void tcp_setup(mongo_link* link){
+  int error, handle;
+  struct hostent *host;
+  struct sockaddr_in server;
+
+  host = gethostbyname (link->master->host);
+  handle = socket (AF_INET, SOCK_STREAM, 0);
+  if (handle == -1){
+    handle = 0;
+  }
+  else {
+    server.sin_family = AF_INET;
+    server.sin_port = htons (link->master->port);
+    server.sin_addr = *((struct in_addr *) host->h_addr);
+    bzero (&(server.sin_zero), 8);
+
+    error = connect(handle, (struct sockaddr *) &server, sizeof (struct sockaddr));
+    if (error == -1){
+      handle = 0;
+    }
+  }
+
+  link->master->socket = handle;
+}
+
+// Disconnect & free connection struct
+void ssl_disconnect (mongo_link *link){
+  if(link->ssl_handle){
+    SSL_shutdown (link->ssl_handle);
+    SSL_free (link->ssl_handle);
+  }
+
+  if (link->ssl_context)
+    SSL_CTX_free (link->ssl_context);
+}
+#endif
